/*    
 *    Copyright (c) 2010 LongTail AdSolutions, Inc
 *
 *    This file is part of the Open Video Ads VAST framework.
 *
 *    The VAST framework is free software: you can redistribute it 
 *    and/or modify it under the terms of the GNU General Public License 
 *    as published by the Free Software Foundation, either version 3 of 
 *    the License, or (at your option) any later version.
 *
 *    The VAST framework is distributed in the hope that it will be 
 *    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with the framework.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.openvideoads.vast.model {
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	
	import org.openvideoads.base.Debuggable;
	import org.openvideoads.regions.view.FlashMedia;
	import org.openvideoads.util.BrowserUtils;
	import org.openvideoads.util.PopupWindow;
	import org.openvideoads.util.StringUtils;
	import org.openvideoads.vast.overlay.OverlayView;
	import org.openvideoads.vast.overlay.VPAIDWrapper;
	import org.openvideoads.vpaid.IVPAID;
	import org.openvideoads.vpaid.VPAIDEvent;
	
	/**
	 * @author Paul Schulz
	 */
	public class VPAIDMediaFile extends MediaFile implements VPAIDPlayback, FlashMedia {
		protected var _vpaidWrapper:VPAIDWrapper = null;
		protected var _vpaidVersion:String = null;
		protected var _onAdCompleteHandler:Function;
		protected var _onAdErrorHandler:Function;
		protected var _onAdStartHandler:Function;
		protected var _onAdLinearChangeHandler:Function;
		protected var _onAdExpandedChangeHandler:Function;
		protected var _onRemainingTimeChangeHandler:Function;
		protected var _onLoadedHandler:Function;
		protected var _onImpressionHandler:Function;
		protected var _onClickThruHandler:Function;
		protected var _onUserAcceptInvitationHandler:Function;
		protected var _onMinimizeHandler:Function;
		protected var _onCloseHandler:Function;
		protected var _onVideoStartHandler:Function;
		protected var _onVideoFirstQuartileHandler:Function;
		protected var _onVideoMidpointHandler:Function;
		protected var _onVideoThirdQuartileHandler:Function;
		protected var _onVideoCompleteHandler:Function;
		protected var _onPausedHandler:Function;
		protected var _onPlayingHandler:Function;
		protected var _externalAPICallHandler:Function = null;
		protected var _timeoutTimer:Timer = null;
		protected var _maxDurationTimeoutInSeconds:int = 65;
		protected var _maxDurationTimeoutEnabled:Boolean = false;
		protected var _activeOverlay:OverlayView = null;
		protected var _running:Boolean = false;
		protected var _enforceRecommendedSizing:Boolean = false;
		protected var _adTerminated:Boolean = false;
		protected var _resourceLoaded:Boolean = false;

		public function VPAIDMediaFile() {
			super();
		}

		public function signalLoadError(errorMessage:String):void {
			callbackOnError(errorMessage)
		}

		public function get recommendedWidth():int {
			if(_width != null) {
				return parseInt(_width);
			}
			return -1;
		}
		
		public function get recommendedHeight():int {
			if(_height != null) {
				return parseInt(_height);
			}
			return -1;
		}

		public function set enforceRecommendedSizing(enforceRecommendedSizing:Boolean):void {
			_enforceRecommendedSizing = enforceRecommendedSizing;
		} 
		
		public function get enforceRecommendedSizing():Boolean {
			return _enforceRecommendedSizing;
		}

		public function markAsLoading():void {
			_resourceLoaded = false;
		}

		public function markAsLoaded():void {
			_resourceLoaded = true;
		}

		public function get loaded():Boolean {
			return _resourceLoaded;
		}

		public function setOverlay(overlay:OverlayView):void {
			_activeOverlay = overlay;
		}
		
		public function getOverlay():OverlayView {
			return _activeOverlay;	
		}
		
		public function hasActiveOverlay():Boolean {
			return (_activeOverlay != null);
		}
		
		public function closeActiveOverlay():void {
			if(hasActiveOverlay()) {
				_activeOverlay.hide();
				_activeOverlay = null;
			}
		}
		
		public override function unload():void {
			CONFIG::debugging { doLog("Unloading the Linear VPAID resource...", Debuggable.DEBUG_VPAID); }
			if(isRunning()) {
				CONFIG::debugging { doLog("The VPAID Ad is still running sending stop() before unloading the resource", Debuggable.DEBUG_VPAID); }
				_adTerminated = true;
				_vpaidWrapper.stopAd();
			}
			removeListeners();
			if(hasActiveOverlay()) {
				_activeOverlay.clearDisplayContent();
				closeActiveOverlay();			
			}
			_vpaidWrapper = null;
		}

		public function getVPAID():IVPAID {
			if(_vpaidWrapper != null) {
				return _vpaidWrapper;
			}
			return null;
		}
		
		public function isRunning():Boolean {
			return (_vpaidWrapper != null) && _running;
		}

		public function isInNonLinearMode():Boolean {
			if(isRunning()) {
				return (_vpaidWrapper.adLinear == false) && (_vpaidWrapper.adExpanded == false);
			}
			return false;
		}

		public function registerStartHandler(onStartHandler:Function):void {	
			_onAdStartHandler = onStartHandler;
		}

		public function registerErrorHandler(onErrorHandler:Function):void {	
			_onAdErrorHandler = onErrorHandler;
		}
		
		public function registerCompleteHandler(onCompleteHandler:Function):void {
			_onAdCompleteHandler = onCompleteHandler;
		}

		public function registerExpandedChangeHandler(onExpandedChangeHandler:Function):void {
			_onAdExpandedChangeHandler = onExpandedChangeHandler;
		}

		public function registerLinearChangeHandler(onLinearChangeHandler:Function):void {
			_onAdLinearChangeHandler = onLinearChangeHandler;
		}

		public function registerRemainingTimeChangeHandler(onRemainingTimeChangeHandler:Function):void {
			_onRemainingTimeChangeHandler = onRemainingTimeChangeHandler;
		}

		public function registerLoadedHandler(onLoadedHandler:Function):void {
			_onLoadedHandler = onLoadedHandler
		}

		public function registerImpressionHandler(onImpressionHandler:Function):void {
			_onImpressionHandler = onImpressionHandler;
		}

		public function registerClickThruHandler(onClickThruHandler:Function):void {
			_onClickThruHandler = onClickThruHandler;	
		}

		public function registerUserAcceptInvitationHandler(onUserAcceptInvitationHandler:Function):void {
			_onUserAcceptInvitationHandler = onUserAcceptInvitationHandler;
		}

		public function registerMinimizeHandler(onMinimizeHandler:Function):void {
			_onMinimizeHandler = onMinimizeHandler;
		}

		public function registerCloseHandler(onCloseHandler:Function):void {
			_onCloseHandler = onCloseHandler;
		}

		public function registerVideoStartHandler(onVideoStartHandler:Function):void {
			_onVideoStartHandler = onVideoStartHandler;
		}

		public function registerVideoFirstQuartileHandler(onVideoFirstQuartileHandler:Function):void {
			_onVideoFirstQuartileHandler = onVideoFirstQuartileHandler;
		}

		public function registerVideoMidpointHandler(onVideoMidpointHandler:Function):void {
			_onVideoMidpointHandler = onVideoMidpointHandler;
		}

		public function registerVideoThirdQuartileHandler(onVideoThirdQuartileHandler:Function):void {
			_onVideoThirdQuartileHandler = onVideoThirdQuartileHandler;
		}

		public function registerVideoCompleteHandler(onVideoCompleteHandler:Function):void {
			_onVideoCompleteHandler = onVideoCompleteHandler;
		}
		
		public function registerOnPausedHandler(onPausedHandler:Function):void {
			_onPausedHandler = onPausedHandler;
		}
		
		public function registerOnPlayingHandler(onPlayingHandler:Function):void {
			_onPlayingHandler = onPlayingHandler;
		}
		
		public function registerExternalAPICallHandler(externalAPICallHandler:Function):void {
			_externalAPICallHandler = externalAPICallHandler;
		}

		public function setMaxDurationTimeout(maxDurationTimeout:int):void {
			_maxDurationTimeoutInSeconds = maxDurationTimeout;
		}

		public function enableMaxDurationTimeout():void {
			_maxDurationTimeoutEnabled = true;
		}

		public function disableMaxDurationTimeout():void {
			_maxDurationTimeoutEnabled = false;
		}
		
		protected function fireExternalAPICall(eventType:String, data:Object=null):void {
			if(_externalAPICallHandler != null) {
				_externalAPICallHandler(new VPAIDEvent(eventType, data));
			}	
		}
		
		protected function callbackOnStart():void {
			_running = true;
			if(_onAdStartHandler != null) {
				_onAdStartHandler(new VPAIDEvent(VPAIDEvent.AdStarted));
			}
		}
		
		protected function callbackOnError(message:String):void {
			_running = false;
			if(_onAdErrorHandler != null) {
				_onAdErrorHandler(new VPAIDEvent(VPAIDEvent.AdError, message));	
			}			
		}
		
		protected function callbackOnComplete(eventType:String):void {
			_running = false;
			if(_onAdCompleteHandler != null) {
				_onAdCompleteHandler(new VPAIDEvent(eventType, { terminated: _adTerminated }));
			}
		}

		protected function callbackOnLoaded():void {
			if(_onLoadedHandler != null) {
				_onLoadedHandler(new VPAIDEvent(VPAIDEvent.AdLoaded));
			}
		}
			
		protected function callbackOnImpression():void {
			if(_onImpressionHandler != null) {
				_onImpressionHandler(new VPAIDEvent(VPAIDEvent.AdImpression));
			}
		}
		
		protected function callbackOnClickThru():void {
			if(_onClickThruHandler != null) {
				_onClickThruHandler(new VPAIDEvent(VPAIDEvent.AdClickThru));
			}			
		}
		
		protected function callbackOnUserAcceptInvitation():void {
			if(_onUserAcceptInvitationHandler != null) {
				_onUserAcceptInvitationHandler(new VPAIDEvent(VPAIDEvent.AdUserAcceptInvitation));
			}			
		}
		
		protected function callbackOnMinimize():void {
			if(_onMinimizeHandler != null) {
				_onMinimizeHandler(new VPAIDEvent(VPAIDEvent.AdUserMinimize));
			}			
		}
		
		protected function callbackOnClose():void {
			if(_onCloseHandler != null) {
				_onCloseHandler(new VPAIDEvent(VPAIDEvent.AdUserClose));
			}			
		}
		
		protected function callbackOnVideoStart():void {
			if(_onVideoStartHandler != null) {
				_onVideoStartHandler(new VPAIDEvent(VPAIDEvent.AdVideoStart));
			}			
		}
		
		protected function callbackOnVideoFirstQuartile():void {
			if(_onVideoFirstQuartileHandler != null) {
				_onVideoFirstQuartileHandler(new VPAIDEvent(VPAIDEvent.AdVideoFirstQuartile));
			}			
		}
		
		protected function callbackOnVideoMidpoint():void {
			if(_onVideoMidpointHandler != null) {
				_onVideoMidpointHandler(new VPAIDEvent(VPAIDEvent.AdVideoMidpoint));
			}
		}
		
		protected function callbackOnVideoThirdQuartile():void {
			if(_onVideoThirdQuartileHandler != null) {
				_onVideoThirdQuartileHandler(new VPAIDEvent(VPAIDEvent.AdVideoThirdQuartile));
			}			
		}
		
		protected function callbackOnVideoComplete():void {
			if(_onVideoCompleteHandler != null) {
				_onVideoCompleteHandler(new VPAIDEvent(VPAIDEvent.AdVideoComplete));
			}			
		}
		
		protected function callbackOnPaused():void {
			if(_onPausedHandler != null) {
				_onPausedHandler(new VPAIDEvent(VPAIDEvent.AdPaused));
			}			
		}
		
		protected function callbackOnPlaying():void {
			if(_onPlayingHandler != null) {
				_onPlayingHandler(new VPAIDEvent(VPAIDEvent.AdPlaying));
			}			
		}

		public function get swfURL():String {
			if(_url != null) {
				return _url.url;			
			}
			return null;
		}
		
		protected function removeListeners():void {
			if(_vpaidWrapper != null) {
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdLoaded, onVPAIDAdLoaded);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdStarted, onVPAIDAdStarted);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdStopped, onVPAIDAdStopped);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdLinearChange, onVPAIDAdLinearChange);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdExpandedChange, onVPAIDAdExpandedChange);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdRemainingTimeChange, onVPAIDAdRemainingTimeChange);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdVolumeChange, onVPAIDAdVolumeChange);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdImpression, onVPAIDAdImpression);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdVideoStart, onVPAIDAdVideoStart);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdVideoFirstQuartile, onVPAIDAdVideoFirstQuartile);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdVideoMidpoint, onVPAIDAdVideoMidpoint);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdVideoThirdQuartile, onVPAIDAdVideoThirdQuartile);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdVideoComplete, onVPAIDAdVideoComplete);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdClickThru, onVPAIDAdClickThru);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdUserAcceptInvitation, onVPAIDAdUserAcceptInvitation);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdUserClose, onVPAIDAdUserClose);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdPaused, onVPAIDAdPaused);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdPlaying, onVPAIDAdPlaying);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdLog, onVPAIDAdLog);
				_vpaidWrapper.removeEventListener(VPAIDEvent.AdError, onVPAIDAdError);				
			}
		}
		
		public function registerAsVPAID(vpaidWrapper:VPAIDWrapper):Boolean {
			if(vpaidWrapper != null) {
				_vpaidWrapper = vpaidWrapper;
                _vpaidVersion = (_vpaidWrapper.isV110() ? "1.1.0" : (_vpaidWrapper.isV100() ? "1.0.0" : null));
				
				// Handshake on the API version

				var vpaidVersionSupportedByAd:String = null;		
						
				if(_vpaidVersion == "1.0.0") {
					CONFIG::debugging { doLog("VPAID 1.0.0 Ad loaded - interface not supported", Debuggable.DEBUG_VPAID); }
					vpaidVersionSupportedByAd = _vpaidWrapper.initVPAIDVersion("1.0.0");
				}
				else if(_vpaidVersion == "1.1.0") {
					vpaidVersionSupportedByAd = _vpaidWrapper.handshakeVersion("1.1.0");
					CONFIG::debugging { doLog("VPAID 1.1.0 Ad loaded - VPAID version handshake reports VPAID Ad version as " + vpaidVersionSupportedByAd, Debuggable.DEBUG_VPAID);	}				
					
					// register all the event handlers
					
					_vpaidWrapper.addEventListener(VPAIDEvent.AdLoaded, onVPAIDAdLoaded);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdStarted, onVPAIDAdStarted);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdStopped, onVPAIDAdStopped);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdLinearChange, onVPAIDAdLinearChange);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdExpandedChange, onVPAIDAdExpandedChange);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdRemainingTimeChange, onVPAIDAdRemainingTimeChange);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdVolumeChange, onVPAIDAdVolumeChange);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdImpression, onVPAIDAdImpression);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdVideoStart, onVPAIDAdVideoStart);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdVideoFirstQuartile, onVPAIDAdVideoFirstQuartile);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdVideoMidpoint, onVPAIDAdVideoMidpoint);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdVideoThirdQuartile, onVPAIDAdVideoThirdQuartile);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdVideoComplete, onVPAIDAdVideoComplete);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdClickThru, onVPAIDAdClickThru);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdUserAcceptInvitation, onVPAIDAdUserAcceptInvitation);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdUserClose, onVPAIDAdUserClose);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdPaused, onVPAIDAdPaused);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdPlaying, onVPAIDAdPlaying);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdLog, onVPAIDAdLog);
					_vpaidWrapper.addEventListener(VPAIDEvent.AdError, onVPAIDAdError);				

					return true;
				}
				else {
					CONFIG::debugging { doLog("VPAID Ad Loaded but it does not have a valid VPAID API version (reporting as '" + _vpaidVersion + "') - ignoring completely", Debuggable.DEBUG_VPAID); }
				}
			}		
			return false;
		}

		public function startVPAID(width:Number=-1, height:Number=-1, mode:String="normal", passReferrer:Boolean=false, referrer:String=null):void {
			if(_vpaidWrapper != null) {
				_adTerminated = false;
				if(_vpaidVersion == "1.1.0") {
					var environmentVars:String = "";
					if(passReferrer) {
						if(referrer != null) {
							environmentVars = "referrer=" + referrer;
							CONFIG::debugging { doLog("Passing in the referrer manually as '" + referrer + "' to the VPAID ad via the environmentVars", Debuggable.DEBUG_VPAID); }
						}
						else {
							try {
								var referrer:String = BrowserUtils.getReferrer(true);
								environmentVars = "referrer=" + referrer;
								CONFIG::debugging { doLog("Passing in the referrer automatically as '" + referrer + "' to the VPAID ad via the environmentVars", Debuggable.DEBUG_VPAID); }
							}
							catch(e:Error) { };
						}
					}
					CONFIG::debugging { doLog("Initialising the VPAID ad with width:" + Math.floor(width) + ", height:" + Math.floor(height) + ", mode:" + mode + ", bitrate:400, parameters:" + adParameters + ", environment vars:" + environmentVars, Debuggable.DEBUG_VPAID); }
					try {
						_vpaidWrapper.initAd(
								Math.floor(width), 
								Math.floor(height), 
								mode, 
								400, 
								(adParameters == null) ? "" : adParameters, 
								(environmentVars == null) ? "" : environmentVars
						);
						startTimeoutTimer();
					}
					catch(e:Error) {
						callbackOnError("Exception generated in VPAID.initAd() - aborting - " + e.message);
					}
				}
				else {
					// version not handled
				}
			}
		}
		
		public function resize(width:Number, height:Number, viewMode:String):void {
			if(_vpaidWrapper != null) {
				_vpaidWrapper.resizeAd(width, height, viewMode);				
			}	
		}		

		public function pause():void {
			if(_vpaidWrapper != null) {
				_vpaidWrapper.pauseAd();
			}				
		}		

		public function resume():void {
			if(_vpaidWrapper != null) {
				_vpaidWrapper.resumeAd();
			}				
		}		

		protected function startTimeoutTimer():void {
			if(_maxDurationTimeoutEnabled) {
				if(_timeoutTimer == null) {
					var seconds:int = ((_parentAd != null) ? ((_parentAd.getDurationAsInt() > 0) ? _parentAd.getDurationAsInt() : _maxDurationTimeoutInSeconds) : _maxDurationTimeoutInSeconds);
					if(seconds > 0) {
						if(_timeoutTimer != null) stopTimeoutTimer();
						_timeoutTimer = new Timer(seconds * 1000, 2);
					    _timeoutTimer.addEventListener(TimerEvent.TIMER, stopTimeoutTimer);
					    _timeoutTimer.start();	
						CONFIG::debugging { doLog("VPAIDMediaFile::startTimeoutTimer() - timer started - timeout set at " + seconds + " seconds", Debuggable.DEBUG_VPAID);	}			
					}
					else {
						CONFIG::debugging { doLog("VPAIDMediaFile::startTimeoutTimer() - cannot set timeout timer - no duration specified for this ad", Debuggable.DEBUG_VPAID);	}	
					}						
				} 
				else {
					CONFIG::debugging { doLog("VPAIDMediaFile::startTimeoutTimer() - ignoring request to set timeout time - timer is already active", Debuggable.DEBUG_VPAID);	}	
				}		
			}
			else {
				CONFIG::debugging { doLog("VPAIDMediaFile::startTimeoutTimer() - not setting a max duration timeout timer - disabled in config", Debuggable.DEBUG_VPAID); }
			}
		}
		
		public function stopTimeoutTimer(e:TimerEvent=null):void {
			if(_timeoutTimer != null) {
				CONFIG::debugging { doLog("VPAIDMediaFile::stopTimeoutTimer() - timer " + ((e == null) ? "aborted - no longer required" : "stopped - timeout triggered"), Debuggable.DEBUG_VPAID);	}
				_timeoutTimer.stop();
				_timeoutTimer = null;
				if(e != null) {
					try {
						_vpaidWrapper.stopAd();
						callbackOnError("Timeout triggered on VPAID Ad startup - also for safety, tried to forcibly stop the VPAID ad before firing this error");
					}
					catch(e:Error) {
						callbackOnError("Timeout timer triggered - forcing stop on exception generated from VPAID.stopAd() - " + e.message);
					}
				}
			}
		}
		
		public override function isInteractive():Boolean {
			return true;
		}
		
		protected function triggerTrackingEvent(eventType:String):void {
			if(_parentAd != null) {
				if(_parentAd.parentAdContainer != null) {
					if(_parentAd.parentAdContainer.hasWrapper()) {
						_parentAd.parentAdContainer.wrapper.triggerTrackingEvent(eventType);
					}
					else _parentAd.parentAdContainer.triggerTrackingEvent(eventType);				
				}
			}
		}
		
		protected function triggerImpressionConfirmations():void {
			if(_parentAd != null) {
				if(_parentAd.parentAdContainer != null) {
					if(_parentAd.parentAdContainer.hasWrapper()) {
						_parentAd.parentAdContainer.wrapper.triggerImpressionConfirmations();
					}
					_parentAd.parentAdContainer.triggerImpressionConfirmations();
				}
			}			
		}
		
		protected function triggerClickTracking():void {
			CONFIG::debugging { doLog("VPAIDMediaFile::triggerClickTracking() - firing click tracking urls", Debuggable.DEBUG_VPAID); }
			if(_parentAd != null) {
				_parentAd.fireClickTracking();				
			}			
		}	
		
        // VPAID Event Handlers
        
		protected function onVPAIDAdLoaded(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLoaded event received - starting Ad", Debuggable.DEBUG_VPAID); }
			try {
				fireExternalAPICall(VPAIDEvent.AdLoaded);
				_vpaidWrapper.startAd();												
				callbackOnLoaded();
			}
			catch(e:Error) {
				callbackOnError("Exception generated in VPAID.startAd() - aborting - " + e.message);
			}
		}		

		protected function onVPAIDAdStarted(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdStarted event received", Debuggable.DEBUG_VPAID); }
			callbackOnStart();
			triggerTrackingEvent(TrackingEvent.EVENT_START);
			fireExternalAPICall(VPAIDEvent.AdStarted);
		}		

		protected function onVPAIDAdStopped(event:*):void {
			stopTimeoutTimer();
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdStopped event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_STOP);
			callbackOnComplete(VPAIDEvent.AdStopped);
		}		

		protected function onVPAIDAdLinearChange(event:*):void {	
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLinearChange event received (adLinear == " + _vpaidWrapper.adLinear + ")", Debuggable.DEBUG_VPAID); }
			if(_onAdLinearChangeHandler != null) {
				_onAdLinearChangeHandler(new VPAIDEvent(VPAIDEvent.AdLinearChange, _vpaidWrapper.adLinear));
			}
		}
		
		protected function onVPAIDAdExpandedChange(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdExpandedChange event received (adExpanded == " + _vpaidWrapper.adExpanded + ")", Debuggable.DEBUG_VPAID); }
			if(_vpaidWrapper.adExpanded) {
				triggerTrackingEvent(TrackingEvent.EVENT_EXPAND);
			}
			else {
				triggerTrackingEvent(TrackingEvent.EVENT_COLLAPSE);				
			}
			if(_onAdExpandedChangeHandler != null) {
				_onAdExpandedChangeHandler(new VPAIDEvent(VPAIDEvent.AdExpandedChange, { expanded: _vpaidWrapper.adExpanded, linearPlayback: _vpaidWrapper.adLinear }));
			}
		}

		protected function onVPAIDAdRemainingTimeChange(event:*):void {
			var remainingTime:Number = _vpaidWrapper.adRemainingTime;
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdRemainingTimeChange event received (adRemainingTime == " + remainingTime + ")", Debuggable.DEBUG_VPAID); }
			if(_onRemainingTimeChangeHandler != null) {
				_onRemainingTimeChangeHandler(new VPAIDEvent(VPAIDEvent.AdRemainingTimeChange, remainingTime));
			}
		}		
		
		protected function onVPAIDAdVolumeChange(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdVolumeChange event received", Debuggable.DEBUG_VPAID); }
			fireExternalAPICall(VPAIDEvent.AdVolumeChange, event.data);
		}		
		
		protected function onVPAIDAdImpression(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdImpression event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_CREATIVE_VIEW);
			triggerImpressionConfirmations();
			fireExternalAPICall(VPAIDEvent.AdImpression);
			callbackOnImpression();
		}		
		
		protected function onVPAIDAdVideoStart(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdVideoStart event received", Debuggable.DEBUG_VPAID); }
			fireExternalAPICall(VPAIDEvent.AdVideoStart);
			callbackOnVideoStart();
		}		
		
		protected function onVPAIDAdVideoFirstQuartile(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdVideoFirstQuartile event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_1STQUARTILE);
			fireExternalAPICall(VPAIDEvent.AdVideoFirstQuartile);
			callbackOnVideoFirstQuartile();
		}		
		
		protected function onVPAIDAdVideoMidpoint(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdVideoMidpoint event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_MIDPOINT);
			fireExternalAPICall(VPAIDEvent.AdVideoMidpoint);
			callbackOnVideoMidpoint();
		}		
		
		protected function onVPAIDAdVideoThirdQuartile(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdVideoThirdQuartile event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_3RDQUARTILE);
			fireExternalAPICall(VPAIDEvent.AdVideoThirdQuartile);
			callbackOnVideoThirdQuartile();
		}		
		
		protected function onVPAIDAdVideoComplete(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdVideoComplete event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_COMPLETE);
			fireExternalAPICall(VPAIDEvent.AdVideoComplete);
			callbackOnVideoComplete();
		}
		
		protected function onVPAIDAdClickThru(event:*):void {
			try {
				// The event can tell the player to handle the processing of the click through URL
				// Event data for handling the click event comes in the form
				// var data:Object = { "playerHandles": true, "url": "http://destination_page" };
				if(event.data != undefined && event.data != null) {
					if(event.data.playerHandles && !StringUtils.isEmpty(event.data.url)) {
						CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdClickThru event received - VPAID ad has requested browser to open click through URL " + event.data.url, Debuggable.DEBUG_VPAID);	}	
						PopupWindow.openWindow(event.data.url, "_blank");
					}
					else {
						CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdClickThru event received - VPAID ad must be handling click through because data.playerHandles == " + event.data.playerHandles + " and data.url == " + event.data.url, Debuggable.DEBUG_VPAID);	}
					}		
				}
				else {
					CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdClickThru event received - VPAID ad internally opening click through URL", Debuggable.DEBUG_VPAID); }
				}
			}
			catch(e:Error) {
			}
			triggerClickTracking();
			fireExternalAPICall(VPAIDEvent.AdClickThru);
			callbackOnClickThru();
		}		
		
		protected function onVPAIDAdUserAcceptInvitation(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdUserAcceptInvitation event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_ACCEPT);
			fireExternalAPICall(VPAIDEvent.AdUserAcceptInvitation);
			callbackOnUserAcceptInvitation();
		}		
		
		protected function onVPAIDAdUserMinimize(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdUserMinimize event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_COLLAPSE);
			fireExternalAPICall(VPAIDEvent.AdUserMinimize);
			callbackOnMinimize();
		}		
		
		protected function onVPAIDAdUserClose(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdUserClose event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_CLOSE);
			fireExternalAPICall(VPAIDEvent.AdUserClose);
			callbackOnClose();
		}		
		
		protected function onVPAIDAdPaused(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdPaused event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_PAUSE);
			stopTimeoutTimer();
			fireExternalAPICall(VPAIDEvent.AdPaused);
			callbackOnPaused();
		}		
		
		protected function onVPAIDAdPlaying(event:*):void {
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdPlaying event received", Debuggable.DEBUG_VPAID); }
			triggerTrackingEvent(TrackingEvent.EVENT_RESUME);
			startTimeoutTimer();
			fireExternalAPICall(VPAIDEvent.AdPlaying);
			callbackOnPlaying();
		}		
		
		protected function onVPAIDAdLog(event:*):void {
			if(event != null) {
                if(event.data != null) {
                	if(event.data is String) {
						CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLog - " + event.data, Debuggable.DEBUG_VPAID); }           		                		
						fireExternalAPICall(VPAIDEvent.AdLog, event.data);
						return;
                	}
                	else if(event.data is Object) {
						CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLog - " + event.data.message, Debuggable.DEBUG_VPAID); }           		
						fireExternalAPICall(VPAIDEvent.AdLog, event.data);
						return;
                	}
					else {
						CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLog - can't work out how to print out the VPAID log message", Debuggable.DEBUG_VPAID); }
					}            		
                }
				else {
					CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLog - no data supplied in log event", Debuggable.DEBUG_VPAID); }
				}
			}
			else {
				CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdLog - no event supplied to logging call", Debuggable.DEBUG_VPAID); }
			}

			fireExternalAPICall(VPAIDEvent.AdLog);
		}		

		protected function onVPAIDAdError(event:*):void {
			stopTimeoutTimer();
			var message:String = ((event.data != null) ? ((event.data.message != undefined) ? event.data.message : event.data) : "no error message provided");
			CONFIG::debugging { doLog("VPAIDMediaFile::VPAID.AdError event received - " + message, Debuggable.DEBUG_VPAID); }
			callbackOnError(message);
		}	
		
		public function toRuntimeStateJSObject():Object {
			if(_vpaidWrapper != null) {
				return _vpaidWrapper.toRuntimeStateJSObject();			
			}
			else return new Object();
		}
	}
}